

(in-package :utilities-testing)


;;; 
;;; First, some comparison implementations.

;;; FILTER-IF unoptimized
(defun filter-if-un (fn list &key (key #'identity))
  (utilities::%filter-if fn list :key key))

(compile 'filter-if-un)


;;; FILTER unoptimized
(defun filter-un (item/filter-set list &key (key #'identity) (test #'equalp) 
		  (filter-set nil) (filter-key #'identity))
  (utilities::%filter item/filter-set list 
		      :key key :test test :filter-set filter-set :filter-key filter-key))

(compile 'filter-un)

 
;;; FILTER-IF-NR/P: nreverse/push
(defun filter-if-nr/p (fn lst &key (key #'identity))
  (let ((acc nil))
    (dolist (x lst)
      (let ((val (funcall fn (funcall key x))))
	(if val (push val acc))))
    (nreverse acc)))

(compile 'filter-if-nr/p)


;;; FILTER-IF-CL: using non-deprecated library functions with non-deprecated arguments
;;; (this functions returns all items and not whatever fn returns but still it's still 
;;; good to compare against) 
(defun filter-if-cl (fn lst &key (key #'identity))
  (remove-if (complement fn)  lst :key key))

(compile 'filter-if-cl)


;;; FILTER-IF-CNK: collect no key: it must be included in fn
(defun filter-if-cnk (fn lst)
  (collecting 
    (dolist (x lst)
      (let ((val (funcall fn x)))
	(when val  ;; how handy awhen would be here...
	  (collect val))))))

(compile 'filter-if-cnk)


;;; FILTER-CL: using non-deprecated CL function with non-deprecated arguments
(defun filter-cl (item lst &key (key #'identity) (test #'equal))
  (filter-if-cl (lambda(x) 
		   (funcall test x item))
		 lst :key key))

(compile 'filter-cl)


;;;
;;; Performance benchmarking tests

(defbenchmark "FILTER-IF"
    ((l-num (collecting
	      (times 10000 (collect (random 10))))))
  (filter-if #'oddp l-num)
  ("FILTER-IF-NR/P" (filter-if-nr/p #'oddp l-num))
  ("FILTER-IF-CL"  (filter-if-cl #'oddp l-num))
  ("FILTER-IF-CNK"  (filter-if-cnk #'oddp l-num))
  ("FILTER-IF-UN"   (filter-if-un #'oddp l-num)))


(defbenchmark "FILTER NUMBER"
    ((l-num (collecting
	      (times 10000 (collect (random 10)))))
     (num (car (freq-extremum l-num))))
  (filter num l-num :test #'=)
  ("FILTER-IF-NR/P" (filter-if-nr/p (lambda(x) (when (= x num) num)) l-num))
  ("FILTER-IF-CL"   (filter-if-cl (lambda(x) (= x num)) l-num))
  ("FILTER-CL"	    (filter-cl num l-num :test #'=))
  ("FILTER-IF-CNK"  (filter-if-cnk (lambda(x) (when (= x num) num)) l-num))
  ("FILTER-UN"	    (filter-un num l-num :test #'=)))


(defbenchmark "FILTER STRING"
    ((l-str (collecting
	      (times 10000 (collect (format nil "~a" (+ 100000 (random 10)))))))
     (str (car (freq-extremum l-str :pop (remove-duplicates l-str :test #'string=)))))
  (filter str l-str :test #'string=)
  ("FILTER-IF-NR/P" (filter-if-nr/p (lambda(x) (when (string= x str) str)) l-str))
  ("FILTER-IF-CL"  (filter-if-cl (lambda(x) (string= x str)) l-str))
  ("FILTER-CL"	    (filter-cl str l-str :test #'string=))
  ("FILTER-IF-CNK"  (filter-if-cnk (lambda(x) (when (string= x str) str)) l-str))
  ("FILTER-UN"	    (filter-un str l-str :test #'string=)))


(defbenchmark "FILTER SET NUMBERS"
    ((l-num (collecting
	      (times 10000 (collect (random 10)))))
     (set '(3 5)))
  (filter set l-num :test #'= :filter-set t)
  ("FILTER-IF-NR/P" (filter-if-nr/p (lambda(x) (when (find x set :test #'=) x)) l-num))
  ("FILTER-IF-CL"   (filter-if-cl  (lambda(x) (when (find x set :test #'=) x)) l-num))
  ("FILTER-IF-CNK"  (filter-if-cnk (lambda(x) (when (find x set :test #'=) x)) l-num))
  ("FILTER-UN"	    (filter-un set l-num :test #'= :filter-set t)))


(run-benchmark! "FILTER-IF" :runs 10000)
(run-benchmark! "FILTER NUMBER" :runs 10000)
(run-benchmark! "FILTER STRING" :runs 10000)
(run-benchmark! "FILTER SET NUMBERS" :runs 10000)

(switch-compiler-optimizations 'filter-if :off)
(switch-compiler-optimizations 'filter :off)

