;;;; -*- Mode: Lisp -*-

(in-package :utilities-testing)


;;; Benchmarking testing code

(defvar *benchmarks* '())


(defstruct (benchmark 
	     (:print-function (lambda(s stream level)
				(declare (ignore level))
				(format stream "<BENCHMARK-TEST ~a>" 
					(benchmark-name s)))))
  name code test-code versions (time -1) data-bindings)


(defstruct (comparison-version
	     (:print-function (lambda(s stream level)
				(declare (ignore level))
				(format stream "<COMPARISON-VERSION ~a>" 
					(comparison-version-name s)))))
  name code (time -1))


(defmacro defbenchmark (name (&rest bindings) bench-code &rest comp-versions)
  `(let* ,bindings
    (register-benchmark ,name (list ,@(mapcar (lambda(bind)
					       `(list ',(car bind) ,(car bind)))
					     bindings))
    ',bench-code
    ', comp-versions)))


(defun run-benchmark! (name &key (runs 1) (stream t))
  (let ((start-time (get-internal-real-time))
	(test (run-benchmark name runs))
	(end-time (get-internal-real-time)))
    (format t "~&~%")
    (report-benchmark test stream)
    (format stream "~&~%Elapsed time: ~,3F secs" (/ (- end-time start-time) 
						    internal-time-units-per-second))))


(defun report-benchmark (test &optional (stream t))
  (let* ((name (benchmark-name test))
	 (t0 (benchmark-time test)))
    (format stream "~&Test Name: '~a' Time: ~,5F secs~%~%" name t0)
    (format stream "~&~20a ~8a Versus%('~a')" "Comparison Version" "Seconds" name)
    (dolist (cv (benchmark-versions test))
      (let ((t1 (comparison-version-time cv))
	    (vname (format nil "'~a'" (comparison-version-name cv))))
	(format stream "~&~20a ~,5F ~7F%"  vname t1 (versus t1 t0))))))


(defun run-benchmark (name &optional (runs 1))
  (let ((b (find name *benchmarks* :key #'benchmark-name :test #'equal)))
    (if b 
	(funcall (compile nil (benchmark-test-code b)) runs)
	(error "Benchmark <~a> not found." name))
    b))


(defun register-benchmark (name data-bindings bench-code comp-versions)
  (setf *benchmarks* 
	(cons (make-benchmark :name name 
			      :code bench-code 
			      :test-code (gen-benchmark-code name 
							     data-bindings 
							     bench-code 
							     comp-versions)
			      :versions (mapcan (lambda(cv) 
						  (unless (eq (car cv) :prepare)
						    (list (create-comparison-version 
							   (car cv)
							   (cadr cv)))))
						comp-versions)
			      :data-bindings data-bindings)
	      (delete name *benchmarks* :key #'benchmark-name :test #'equal))))


(defun gen-benchmark-code (name data-bindings bench-code comp-versions)
  `(lambda(runs)
    (let* ((b (find ,name *benchmarks* 
		     :key #'benchmark-name :test #'equal))
	   ,@(mapcar (lambda(bind)
		       (let ((var (car bind)))
			 (list var `(cadr 
				     (assoc ',var 
				      (benchmark-data-bindings b))))))
		     data-bindings))
      (setf (benchmark-time b) 
	    (with-runs-measure-time runs ,bench-code))
      ,@(mapcar (lambda(cv)
		  (if (eq (car cv) :prepare)
		      (cadr cv)
		      `(setf (comparison-version-time (find ,(car cv) 
							   (benchmark-versions b)
							   :test #'equal 
							   :key #'comparison-version-name))
			    (with-runs-measure-time runs ,(cadr cv)))))
		comp-versions))))


(defmacro with-runs-measure-time (n &body run-body) ;(&body body)
  (let ((timings (gensym "timings"))
	(t0 (gensym "t0")))
    `(let ((,timings (collecting 
		       (times ,n
			 (let ((,t0 (get-internal-real-time)))
			   ,@run-body
			   (collect (/ (- (get-internal-real-time) ,t0)
				       internal-time-units-per-second)))))))
      ;; extract 5% highest and 5% lowest scores
      (/ (reduce #'+ (delete (n-most-extreme (round (* 0.05 ,n)) ,timings #'>=)
			     (delete (n-most-extreme (round (* 0.05 ,n)) ,timings #'<=) 
				     ,timings)))
       (* 0.9 ,n)))))


(defun create-comparison-version (name comp-code)
  (make-comparison-version :name name  :code comp-code))


(defun remove-benchmark (name)
  (setf *benchmarks* (delete name *benchmarks* :key #'benchmark-name :test #'equal)))


(defun clear-benchmarks (&rest names)
  (if names
      (mapc #'remove-benchmark names)
      (setf *benchmarks* nil)))


;;; 
;;; Useful general utilities for testing.

(defun versus (v base)
  "Percentage difference between V and BASE in terms of BASE."
   (round (* (/ (- v  base) base)
	     100)))

(defun freq-i (i seq &key (test #'equal))
  "Compute I's frequency in SEQ"
  (* 100 (/ (float (count i seq :test test :key #'identity)) (length seq))))


(defun freq-extremum (seq &key (pred #'<) (pop '(0 1 2 3 4 5 6 7 8 9)))
  "Given a POPulation of elements, return its most frequent element in SEQ as determined
  by PRED."
  (extremum (map 'list #'list  
		 pop (map 'list (lambda(p)  (freq-i p seq)) pop))
	    pred :key #'cadr))



;;;; eof


