;;;;
;;;; General definitions for compile time analysis.

(in-package :utilities)

;;;
;;; The following functions are designed to analyze code at macro-expansion/compile time. 
;;; Note that at compile time 'S is really the CONS '(QUOTE S). And function #'ODDP 
;;; is represented by '(FUNCTION ODDP>). Runtime variable S is compile-time symbol 'S.
;;; So the functions deal with things that will be something at runtime.
;;;
;;; Before proceeding (hmmm...seems like I'm writing a book), a couple of definitions are
;;; helpful:
;;;
;;; "literal adj. (of an object) referenced directly in a program rather than being computed by 
;;; the program; that is, appearing as data in a quote form, or, if the object is a 
;;; self-evaluating object, appearing as unquoted data. ``In the form (cons "one" '("two")), 
;;; the expressions "one", ("two"), and "two" are literal objects.'' "
;;;
;;; "function designator n. - a designator for a function; that is, an object that denotes a 
;;; function and that is one of: a symbol (denoting the function named by that symbol in the 
;;; global environment), or a function (denoting itself).(...)" (FD)
;;;
;;; At compile time a literal FD is something that is already a function (#<FUNCTION ODDP>) 
;;; or will be a literal function (eg, '(QUOTE ODDP) and '(FUNCTION ODDP). 


(defun literal-value-p (x)
  "If X is a literal value. Ex: (QUOTE S), ''A'', 1, NIL,.... 
  The runtime symbol 'S, at compile-time, is actually the CONS '(QUOTE S) and runtime
  variable S is the symbol 'S at compile-time."
  (or (and (atom x) (not (variable-p x)))
      (literal-symbol-p x)
      (literal-cons-p x)
      (literal-function-p x)))


(defun variable-p (x)
  "If X is a variable. At compile-time a variable is non-nil symbol."
  (and (symbolp x) (not (eq x nil))))


(defun literal-symbol-p (x)
  "If X is a literal symbol. Ex: (QUOTE S), NIL, (QUOTE NIL)"
  (or (eql x nil)
      (and (consp x) 
	   (eql (first x) 'quote) 
	   (symbolp (second x)))))


(defun literal-cons-p (x)
  "If X is a literal CONS. Ex: (QUOTE (1 2 3))"
  (and (consp x) 
       (eql (first x) 'quote)
       (consp (second x))))


(defun literal-function-designator-p (x)
  "Whether X represents a literal function designator, which can be either a literal function
  or a symbol representing the function's name."
  (or (literal-symbol-p x)
      (literal-function-p x)))


(defun literal-function-p (x)
  "Whether X represents a literal function call. Ex:: #'ODDP, '#'ODDP,
  #'(LAMBDA...), '#'(LAMBDA...), (LAMBDA...), '(LAMBDA...) (see LAMBDA-P).

  Note: This function is designed to work at compile time and, as such,
  a macro parameter that's supposed to, say, hold the function object #'oddp will 
  actually hold the cons '(function +) at compile-time (read macros were already
  expanded). There is one exception though: optional/keyword parameters and their default 
  values. If the macro parameter in the example is an optional parameter then it _will_ 
  hold the function object given by #'oddp."
  (or (sharp-quoted-p x)
      (lambda-p x)
      (and (atom x) (functionp x))))


(defun sharp-quoted-p (x)
  "Whether X is a sharp-quoted form like '#'ODDP."
  (and (consp x)
       (eql (first x) 'function)
       (or (symbolp (second x))
	   (lambda-p (second x)))))


(defun lambda-p (x)
  "Whether X is a '(LAMBDA(args) ...) form. lambda is defined as macro that expands
  to #'(LAMBDA(..) at compile time. So it's possible to skip the sharp-quote and just use
  the unexpanded form."
  (and (consp x)
       (eql (first x) 'lambda)
       (consp (second x))))
  
(defun identity-function-p (fn)
  "..."
  (or (equal fn '#'identity)
      (eql fn #'identity)))
	 
;;;
;;; Compiler optimizations management.

(defgeneric switch-compiler-optimizations (fn-design action)
  (:documentation "Mechanism for enabling/disabling compile time optimizations useful for 
  debugging purposes (eg., tracing).")

  (:method ((fn-design symbol) (action (eql :off)))
    (warn "Switching OFF compile time optimizations on '~a'" fn-design)
    (eval `(define-compiler-macro ,fn-design (&whole whole &rest dontcare)
	    (declare (ignore dontcare))
	    whole)))

  (:method ((fn-design symbol) (action (eql :on)))
    (warn "There are NO switchable compile time optimizations for '~a'" fn-design)))


(defmacro define-switchable-compiler-macro (name (&rest compiler-macro-lambda-list) 
							      &body compiler-macro-body)
  "A convenient way to define a compiler macro that can be switched off/on for some purpose."
  `(progn
     (warn "Switching ON compile time optimizations on '~a'" ',name)
     (define-compiler-macro ,name ,compiler-macro-lambda-list
	,@compiler-macro-body)

     (defmethod switch-compiler-optimizations ((fn (eql ',name)) (action (eql :on)))
       (warn "Switching ON compile time optimizations on '~a'" ',name)
       (define-compiler-macro ,name ,compiler-macro-lambda-list
	 ,@compiler-macro-body))))


;;;; eof