

(in-package :utilities)

;;; 
;;; Filter

;;;
;;; Ideal implementation. With no performance/compatibility concerns in mind, <filter> could 
;;; be implemented like this:
;;;
;;; (defgeneric filter (criteria list &key &allow-other-keys)
;;;   (:documentation "Filter LIST using CRITERIA as the filter. All filtered elements are 
;;;   returned and only these ones. ")
;;; 
;;;   (:method ((fn function) (list list) &key (key #'identity))
;;;     "Return all non-nil elements returned by the filtering function FN."
;;;      (collecting
;;;        (dolist (x list)
;;;	    (let ((val (funcall fn (funcall key x))))
;;;	      (when val 
;;;		(collect val))))))
;;; 
;;;   (:method (item (list list) &key (key #'identity) (test #'equalp)) 
;;;     "Same as FIND but returns all items that match item. Same as REMOVE-IF-NOT (which,
;;;     accordingly to CLHS, is deprecated)."
;;;      (filter (lambda(obj)
;;; 	       (when (funcall test item (funcall key obj)) obj))
;;; 	     list))
;;;   
;;;   (:method ((set list) (list list) &key (skey #'identity) (key #'identity) (test #'equalp))
;;;     "All items from SET will be filtered from LIST."
;;;     (mapcan (lambda(item)
;;; 	      (filter (funcall skey item) list :test test :key key))
;;; 	    set)))
;;;
;;; But because this violates the idiom defined for similar CL functions (find, find-if,...) 
;;; and because CLOS introduces some performance penalties regular defuns will be adopted. 


(defmacro %filter-if (fn list &key (key '#'identity))  
  "Just collects whatever FN returns from LIST.
  Development of the filtering algorithm is centralized in this macro since we want to use 
  the same code both in the functions definitions and in the compiler macros expansions."
  (once-only-calls (fn key)
    (once-only (list)
      `(collecting	 
	(dolist (x ,list)
	  (let ((ret-val ,(fn (key 'x))))
	    (when ret-val 
	      (collect ret-val))))))))


(defun filter-if (fn list &key (key #'identity))
  "Filter LIST using function FN returning whatever FN returns."
  (%filter-if fn list :key key))


(define-switchable-compiler-macro filter-if (&whole whole fn list &key (key #'identity))
  "Optimization for the cases in which FN and KEY are literal functions."
  (declare (ignore whole))
  `(%filter-if ,fn ,list :key ,key))


(eval-when (:compile-toplevel :load-toplevel :execute)

  (defun %filter-item (item list &key (key '#'identity) (test '#'equal))
    "Generates code to FILTER all ITEMs from LIST."
    (once-only-calls (test key)
      (once-only (item)
	(with-unique-names (obj)
	  `(%filter-if (lambda(,obj) 
			 (when ,(test item (key obj))
			   ,obj))
	    ,list)))))


  (defun %filter-items (filter-set list &key (key '#'identity) (test '#'equal)
			(filter-key '#'identity))
    "Generates code to FILTER all FILTER-SET element from LIST."
    (once-only-calls (filter-key)
      (with-unique-names (obj)
	`(mapcan (lambda(,obj)
		   ,(%filter-item (filter-key obj) list :test test :key key))
	  ,filter-set)))))


(defmacro %filter  (item/filter-set list &key (key '#'identity) (test '#'equal)
		    (filter-set nil) (filter-key #'identity))
  `(if ,filter-set
      ,(%filter-items item/filter-set list :key key :test test :filter-key filter-key)
      ,(%filter-item item/filter-set list :key key :test test)))


(defun filter (item/filter-set list &key (key #'identity) (test #'equal) 
	       (filter-set nil) (filter-key #'identity))
  "Filter ITEM/FILTER-SET from LIST. Similar to FIND but all 'found' items are 
  returned.
  If ITEM/FILTER-SET is a list, FILTER can either filter each of its elements or just consider 
  that list to be an item.  This behaviour is controlled by FILTER-SET being set to T or not. 
  Note that the first behaviour is similar to INTERSECTION but one can use key functions for 
  both lists."
  (%filter item/filter-set list :key key :test test :filter-set filter-set :filter-key filter-key))


(define-switchable-compiler-macro filter (&whole whole item/filter-set list &key 
						 (key #'identity) (test #'equal)
						 (filter-set nil) (filter-key #'identity))
  "Optimization for the cases in which KEY and TEST are literal functions. Parameter 
  FILTER-SET may also be known at compile-time and so testing for it can be avoided."
  (declare (ignore whole))
  (cond ((eq filter-set t)
	 (%filter-items item/filter-set list :key key :test test :filter-key filter-key))
	((eq filter-set nil)
	 (%filter-item item/filter-set list :key key :test test))
	(t `(%filter ,item/filter-set ,list :key ,key :test ,test 
		     :filter-set ,filter-set :filter-key ,filter-key))))

  
