;;;;
;;;; <once-only-calls>
;;;; Might be useful to read the file <compilation-utilities>.

(in-package :utilities)

;; copied from 'once-only'

(defun %check-once-only-names (names)
  "Check that all of the NAMES are symbols. If not, raise an error."
  ;; This only raises an error for the first non-symbol argument
  ;; found. While this won't report multiple errors, it is probably
  ;; more convenient to only report one.
  (let ((bad-name (find-if-not #'symbolp names)))
    (when bad-name
      (error "ONCE-ONLY-CALLS expected a symbol but got ~S" bad-name))))


(defmacro once-only-calls ((&rest fns) &body body)
  "This macro, like ONCE-ONLY, is useful to write other macros that need to manipulate function
  calls in a safe (meaning, single evaluation) way, yet efficient in those cases that information
  about the functions, like if they're literal functions, is known in advance.

  Given the fn designators in FNS, check for non-literal function designators (FD) - 
  variables or function calls that return functions - and bind each to 'gensymed' variables
  to prevent multiple evaluation. Calls to literal FDs will be optimized by using 
  direct function calls instead of indirect ones via FUNCALL. 
  If any of the FDs is #'IDENTITY then not even the direct function call is generated.

  A local function bounded to each symbol in FNS is created and should be used by the macro
  writer to invoke the respective function."
  (%check-once-only-names fns)
  `(multiple-value-bind (once-only-fn-decls base-calls) (prep-once-only-calls ,@fns)
    (labels ,(mapcar (lambda(fn)
		       `(,fn (&rest args)
			 (apply #'gen-once-only-call base-calls ,fn args)))
		     fns)
      ,(append ``(let ,once-only-fn-decls)
	       body))))

;;;
;;; At compile-time, an FD is anything that represents a function or will represent the
;;; a function at runtime : <'f>, <'#'f>, <#'f>, <(function-that-returns-f)>, 
;;; <variable-or-symbol-macro-bounded-to-function-f>, <#'(lambda)>, <'#'(lambda)> and 
;;; <(lambda)>.

(defun prep-once-only-calls (&rest fns)
  "Prepare a once-only call for each FD in FNS. Returns two lists: the first with declarations 
  for each non-literal function designator to prevent multiple-evaluation. The second is an 
  ASSOC list, keyed by the original function designators, with the datum holding a safe and
  effiecient call to the FNS."
  (let ((once-only-decls nil)
	(base-calls nil))
    (dolist (fn fns (values once-only-decls base-calls))
      (multiple-value-bind (once-only-decl base-call) (prep-once-only-call fn)
	(when once-only-decl
	  (push once-only-decl once-only-decls))
	(push (cons fn base-call) base-calls)))))
	

(defun prep-once-only-call (fn)
  "If FN is a literal FD this function avoids funcalling it. If KEY represents identity 
  then not even the direct function call is generated. Otherwise, a 'gensymed' variable that
  will be bounded to the FN and a funcall to that variable will be generated."
  (cond ((identity-function-p fn)
	 (values nil nil))
	((lambda-p fn)
	 (values nil `(,fn)))
	((functionp fn)
	 (values nil `(funcall ,fn)))
	((literal-function-designator-p fn)
	 (values nil `(,(cadr fn))))
	(t (let ((fn-gensym (gensym)))
	     (values `(,fn-gensym ,fn) `(funcall ,fn-gensym))))))


(defun gen-once-only-call (base-calls fn &rest args)
  "Return a complete call for function designator FN, expanded with the arguments in ARGS.
   The base call (eg, (FUNCALL A) or (CAR)) is retrieved from BASE-CALLS. If no BASE-CALL 
  is found, the arguments in ARGS are returned."
  (let ((base-call (cdr (assoc fn base-calls :test #'equal))))
    (if base-call
      (nconc base-call args)
      (apply #'values args))))


;;;; eof
